//
// Copyright 2021, Jesús Pérez Lorenzo
//
use serde::{Serialize,Deserialize,Deserializer};
use std::fmt;

#[derive(Clone, Debug, Deserialize)]
pub struct Notifier {
  pub name: String,
  pub settings: String,
  pub command: String,
}

#[derive(Clone, Debug, Deserialize, Default)]
pub struct StoreSettings {
  host: String,
  port: u32,
  user: String,
  pass: String,
  pub database: String,
  pub max_conn: u32,
}
impl StoreSettings {
  #[must_use]
  #[allow(unused_variables)]
  pub fn url(&self, store: &str, key: &str) -> String {
    let mut user_pass = String::from("");
    let user: String;
    let pass: String;
    user = self.user.to_owned();
    pass = self.pass.to_owned();
    if user.as_str() != "" || pass.as_str() != "" {
      user_pass = format!("{}:{}", &user, &pass);
    }
    let host = &self.host;
    let port = &self.port;
    let database = &self.database;
    format!(
      "{}://{}@{}:{}/{}",
      store, &user_pass, &host, &port, &database
    )
  }
}
#[derive(Clone, Debug, Deserialize, Default)]
pub struct StoreKeyValue {
  host: String,
  port: u32,
  pub prefix: String,
  pub max_conn: u32,
}
impl StoreKeyValue {
  #[must_use]
  pub fn url(&self, store: &str) -> String {
    let host = &self.host;
    let port = &self.port;
    format!("{}://{}:{}", store, &host, &port)
  }
}
#[derive(Clone, Debug, Deserialize, Default)]
pub struct StoreLocal {
  pub database: String,
}
impl StoreLocal {
  #[must_use]
  pub fn url(&self, store: &str) -> String {
    format!("{}.{}", store, &self.database)
  }
}
#[derive(Clone, Debug, Deserialize, Default)]
pub struct WebServer {
  pub name: String,
	pub home_path: String,
	pub resources_path: String,

  pub certs_store_path: String,
  pub cert_file_sufix: String,

  pub default_module: String,
  // Some paths 
  pub templates_path: String,
  pub defaults_path: String,
  pub dist_path: String,
  pub html_path: String,
  pub upload_path: String,
  /// allow origin for localhost 
  pub allow_origin: Vec<String>,
	/// graphql schemas
  pub gql_schemas_path: String,
  /// graphql query targets
  pub gql_targets: String,
  /// graphql path for requests
  pub gql_req_path: String, 
  /// graphQL path for request
  pub giql_req_path: String,
  pub auth_model_path: String,
  pub auth_policy_path: String,
  pub usrs_store: String,
  pub usrs_store_target: String,
  pub usrs_shadow_store: String,
  pub usrs_shadow_target: String,
  pub main_store: String,
  pub srv_host: String,
  pub srv_port: u16,
  pub srv_protocol: String,
	pub langs: Vec<String>,
  pub signup_mode: String,
  pub password_rules: String,
  pub mapped_url_prefix: String,
  pub admin_key: String,
}
impl WebServer {
  #[must_use]
  pub fn st_html_path(&self) -> &'static str  {
    Box::leak(self.html_path.to_owned().into_boxed_str())
  }
  #[must_use]
  pub fn st_auth_model_path(&self) -> &'static str  {
    Box::leak(self.auth_model_path.to_owned().into_boxed_str())
  }
  #[must_use]
  pub fn st_auth_policy_path(&self) -> &'static str  {
    Box::leak(self.auth_policy_path.to_owned().into_boxed_str())
  }
  #[must_use]
  pub fn st_gql_req_path(&self) -> &'static str  {
    Box::leak(self.gql_req_path.to_owned().into_boxed_str())
  }
  #[must_use]
  pub fn st_giql_req_path(&self) -> &'static str  {
    Box::leak(self.giql_req_path.to_owned().into_boxed_str())
  }
}

#[derive(Eq, PartialEq, Clone, Serialize, Debug, Deserialize)]
pub enum SchedTaskStore {
  None,
  File(String),
  Acid(String),
}
impl Default for SchedTaskStore {
	fn default() -> Self {
		SchedTaskStore::None
	}
}
impl fmt::Display for SchedTaskStore {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		match self {
			SchedTaskStore::None => write!(f,"none"),
			SchedTaskStore::File(p) => write!(f,"file: {}",p),
			SchedTaskStore::Acid(p) => write!(f,"acid: {}",p),
		}
		//write!(f, "{:?}", self)
	}
}

impl SchedTaskStore {
	pub fn get_value(&self) -> String {
		match self {
			SchedTaskStore::None => format!(""),
			SchedTaskStore::File(p) => format!("{}",p),
			SchedTaskStore::Acid(p) => format!("{}",p),
		}
	}
}

fn deserialize_schedtaskstore<'de, D>(deserializer: D) -> Result<SchedTaskStore, D::Error>
where D: Deserializer<'de> {
	let buf = String::deserialize(deserializer)?;
	let v: Vec<&str> = buf.split(':').collect();
	if v.len() > 1 {
		Ok(match v[0] {
			"File" => SchedTaskStore::File(v[1].to_owned()),
			"Acid" => SchedTaskStore::Acid(v[1].to_owned()),
			_ => SchedTaskStore::None
		})
	} else {
		Ok(SchedTaskStore::None)
	}
}

#[derive(Clone, Debug, Deserialize, Default)]
pub struct SchedTask {
  pub name: String,
	#[serde(deserialize_with = "deserialize_schedtaskstore")]
  pub store: SchedTaskStore,
  pub schedule: String,
  pub on_start: bool,
}

#[derive(Clone, Debug, Deserialize, Default)]
pub struct Config {
  pub run_mode: String,
  // warp log name
  pub log_name: String,
  pub cache_path: String,
  pub cache_lock_path: String,
  pub cache_lock_ext: String,
  pub check_path: String,
	pub monitor_rules_path: String,
	pub monitor_rules_file: String,
	pub monitor_rules_format: String,
  pub default_lang: String,
  pub logs_store: String,
  pub logs_format: String,
  pub run_schedtasks: bool,
	pub schedtasks: Vec<SchedTask>,
  pub run_websrvrs: bool,
  pub websrvrs: Vec<WebServer>,
}

impl Config {
  pub fn load_file_content(verbose: &str, cfg_path: &str) -> String {
    let config_path: String;
		if cfg_path.is_empty() {
		 	config_path = envmnt::get_or("APP_CONFIG_PATH", "config.toml");
		} else {
		  config_path = cfg_path.to_string();
		}
		if verbose != "quiet" {
    	println!("Config path: {}", &config_path);
		}
    std::fs::read_to_string(&config_path)
      .unwrap_or_else(|e| {
        eprintln!("Error read {}: {}",&config_path,e);
        String::from("")
    })
  }
  pub fn new(content: String,verbose: &str) -> Self { 
    match toml::from_str(&content) {
      Ok(cfg) => {
				if verbose != "quiet" {
          println!("Config Loaded successfully");
				}
				let app_home=envmnt::get_or("APP_HOME", ".");
				let mut app_cfg: Config = cfg;

				app_cfg.cache_path=format!("{}/{}",&app_home,&app_cfg.cache_path);
				app_cfg.cache_lock_path=format!("{}/{}",&app_home,&app_cfg.cache_lock_path);
				app_cfg.check_path=format!("{}/{}",&app_home,&app_cfg.check_path);
				app_cfg.monitor_rules_path=format!("{}/{}",&app_home,&app_cfg.monitor_rules_path);

				// let mut schedtasks: Vec<SchedTask> = Vec::new();
				// app_cfg.schedtasks.iter().enumerate().for_each(|(pos,it)| {
				// 	schedtasks.push(it.to_owned());
				// 	if !it.path.is_empty() {
				// 		schedtasks[pos].path=format!("{}/{}",&app_home,&it.path)
				// 	}
				// 	if !it.lock_path.is_empty() {
				// 		schedtasks[pos].lock_path=format!("{}/{}",&app_home,&it.lock_path)
				// 	}
				// });

				let mut websrvrs: Vec<WebServer> = Vec::new();

				app_cfg.websrvrs.iter().enumerate().for_each(|(pos,it)| {
					let app_home_path=format!("{}/{}/",&app_home,&it.home_path);
					websrvrs.push(it.to_owned());
					websrvrs[pos].certs_store_path=format!("{}{}",&app_home_path,&it.certs_store_path);
					websrvrs[pos].resources_path=format!("{}{}",&app_home_path,&it.resources_path);
					websrvrs[pos].templates_path=format!("{}{}",&app_home_path,&it.templates_path);
					websrvrs[pos].defaults_path=format!("{}{}",&app_home_path,&it.defaults_path);
					websrvrs[pos].html_path=format!("{}{}",&app_home_path,&it.html_path);
					websrvrs[pos].dist_path=format!("{}{}",&app_home_path,&it.dist_path);
					websrvrs[pos].upload_path=format!("{}{}",&app_home_path,&it.upload_path);
					websrvrs[pos].auth_model_path=format!("{}{}",&app_home_path,&it.auth_model_path);
					websrvrs[pos].auth_policy_path=format!("{}{}",&app_home_path,&it.auth_policy_path);
					websrvrs[pos].usrs_store_target=format!("{}{}",&app_home_path,&it.usrs_store_target);
					websrvrs[pos].usrs_shadow_target=format!("{}{}",&app_home_path,&it.usrs_shadow_target);
				});
				app_cfg.websrvrs = websrvrs;

				app_cfg
      },
      Err(e) => {
        println!("Config error: {}",e);
        Config::default()
      }
    }
  }
	#[must_use]
  pub fn st_log_name(&self) -> &'static str  {
    Box::leak(self.log_name.to_owned().into_boxed_str())
  } 
	pub fn get_schedtask(&self,name: &str) -> SchedTask {
		let filter: Vec<&SchedTask> = self.schedtasks.iter().filter(|it| it.name.as_str() == name).collect();
		if filter.len() > 0 {
			return filter[0].to_owned()
		}
		SchedTask::default()
	}
	pub fn get_websrvr(&self,name: &str) -> WebServer {
		let filter: Vec<&WebServer> = self.websrvrs.iter().filter(|it| it.name.as_str() == name).collect();
		if filter.len() > 0 {
			return filter[0].to_owned()
		}
		WebServer::default()
	}
}
